
package require math::linearalgebra
package require math::constants
package provide compass_utils::math 1.1
namespace eval ::math::linearalgebra { namespace export transpose }
namespace eval m {
    namespace import -force ::math::linearalgebra::*
    ::math::constants::constants pi degtorad radtodeg
    proc vec_angle { vect1 vect2 } {
	set dp [dotproduct $vect1 $vect2]
	set n1 [norm $vect1]
	set n2 [norm $vect2]
	
	if { $n1 == 0.0 || $n2 == 0.0 } {
	    return -code error "Angle not defined for null vector"
	}
	return [expr {acos($dp/$n1/$n2)}]
    }
    proc ellipse_perimeter { rx ry } {
	# Hudson formula (approximate)
	variable pi
	set L [expr {pow($rx-$ry,2)/pow(2.0*($rx+$ry),2)}]
	return [expr {$pi*($rx+$ry)/4.0*(3*(1+$L)+1.0/(1-$L))}]
    }
    proc ellipse_val { c rx ry theta } {
	set x [expr {$rx*cos($theta)}]
	set y [expr {$ry*sin($theta)}]
	return [m::add [list $x $y] $c]
    }
    proc ellipse_normal { rx ry theta } {
	set tx [expr {-1*$rx*sin($theta)}]
	set ty [expr {$ry*cos($theta)}]
	return [m::unitLengthVector [list $ty [expr {-1*$tx}]]]
    }
    proc ellipse_val_inv { c rx ry p } {
	variable pi
	set p [m::sub $p $c]
	set theta $pi
	for { set i 0 } { $i < 100 } { incr i } {
	    set sin [expr {sin($theta)}]
	    set cos [expr {cos($theta)}]
	    set d2 [expr {pow($rx*$cos-[lindex $p 0],2)+pow($ry*$sin-[lindex $p 1],2)}]
	    if { abs($d2) < 1e-7 } { break }
	    set der_d2 [expr {-2*($rx*$cos-[lindex $p 0])*$rx*$sin+
		    2*($ry*$sin-[lindex $p 1])*$ry*$cos}]
	    if { $der_d2 == 0 } { break }
	    set theta [expr {$theta-$d2/$der_d2}]
	}
	while { $theta < 0 } { set theta [expr {$theta+2*$pi}] }
	while { $theta > 2*$pi } { set theta [expr {$theta-2*$pi}] }
	return $theta
    }
    proc ellipse_add_length_to_angle { rx ry theta0 len } {
	# approximating as a circle (very bad)
	set r [expr {.5*($rx+$ry)}]
	set delta_theta [expr {$len/$r}]
	return [expr {$theta0+$delta_theta}]
    }
    proc ellipse_arc_length { rx ry theta1 theta2 { delta .01 } } {
	# from theta1 to theta2 in anti-clockwise orientation
	variable pi
	set len 0.0
	set last_p ""
	while { $theta2 < $theta1 } { set theta2 [expr {$theta2+2*$pi}] }
	for { set t $theta1 } { $t < $theta2 } { set t [expr {$t+$delta}] } {
	    set p [ellipse_val "0 0" $rx $ry $t]
	    if { $last_p ne "" } {
		set len [expr {$len+[m::norm [m::sub $p $last_p]]}]
	    }
	    set last_p $p
	}
	return $len
    }
}
